@@ -51,6 +51,7 @@ from enum import Enum
 import numpy as np
 import numpy.typing as npt
 from attr import Factory, cmp_using, define, field, setters
+from mcpid.lookup import PdgRecords
 from numpy.lib import recfunctions as rfn
 from rich.console import Console
 from rich.tree import Tree
@@ -80,6 +81,8 @@ __all__ = [
 # SET UP ARRAY ATTRIBUTES FOR DATACLASSES #
 ###########################################
 _types = Types()
+_LOOKUP_TABLE = PdgRecords()
+
 DataType = ty.TypeVar("DataType", bound=base.ArrayBase)
 FuncType = ty.TypeVar("FuncType", bound=ty.Callable[..., ty.Any])
 EdgeLike = ty.Union[
@@ -153,10 +156,36 @@ def _array_ufunc(
     **kwargs: ty.Dict[str, ty.Any],
 ) -> ty.Optional[
     ty.Union[
-        ty.Tuple[ty.Union[DataType, base.AnyVector]],
-        ty.Union[DataType, base.AnyVector],
+        ty.Tuple[ty.Union[DataType, base.AnyVector, "MaskArray"], ...],
+        ty.Union[DataType, base.AnyVector, "MaskArray"],
     ]
 ]:
+    """Defines the behaviour of ``ArrayBase`` objects when passed to a
+    numpy ufunc.
+
+    Parameters
+    ----------
+    instance : ArrayBase
+        Array object defined by graphicle.
+    ufunc : callable
+        Numpy ufunc being called on ``instance``.
+    method : str
+        Method applied to ``ufunc``, eg. 'reduce' for ``np.add.reduce``,
+        of which ``np.sum`` is an alias.
+    *inputs : tuple[Any]
+        Positional arguments of the ufunc.
+    **kwargs : dict[str, Any]
+        Keyword arguments of the ufunc.
+
+    Returns
+    -------
+    output : arrays or scalars, or tuple thereof, or None
+        Output of passed ufunc. Type depends on ``method``, ``inputs``,
+        and ``instance``. If ``method`` is 'reduce' on ``instance``
+        with a flat underlying array, then ``output`` will be scalar.
+        If ``ufunc`` maps an array to booleans, ``output`` will be a
+        ``MaskArray``.
+    """
     out = kwargs.get("out", ())
     class_type = instance.__class__
     for x in inputs + out:  # type: ignore
@@ -193,6 +222,9 @@ def _array_ufunc(
 def _array_eq_prep(
     instance: base.ArrayBase, other: ty.Union[base.ArrayBase, base.AnyVector]
 ) -> ty.Tuple[base.AnyVector, base.AnyVector]:
+    """Ensures that the two objects being prepared for comparison are
+    compatible, enabling numpy's vectorised equality operation.
+    """
     other_data = other
     dtype = instance.data.dtype
     if isinstance(other, base.ArrayBase):
@@ -228,6 +260,9 @@ def _array_ne(
 
 
 def _array_repr(instance: base.ArrayBase) -> str:
+    """Provides a common string representation for ``ArrayBase``
+    implementations.
+    """
     data_str = str(instance._data)
     data_splits = data_str.split("\n")
     first_str = data_splits.pop(0)
@@ -268,6 +303,10 @@ def array_field(type_name: str):
 
 
 def _reorder_pmu(array: base.VoidVector) -> base.VoidVector:
+    """Ensures passed structured arrays of 4-momenta have their fields
+    mapped in the same order as graphicle's ``MomentumArray`` underlying
+    convention.
+    """
     names = array.dtype.names
     assert names is not None
     if names == _MOMENTUM_ORDER:
@@ -277,9 +316,52 @@ def _reorder_pmu(array: base.VoidVector) -> base.VoidVector:
     return array[name_reorder]
 
 
-def _array_field(dtype: npt.DTypeLike, num_cols: int = 1, repr: bool = True):
+def _row_contiguous(func: ty.Callable[..., base.AnyVector]):
+    """Decorator to ensure that numpy arrays returned from functions are
+    row-contiguous.
+
+    Parameters
+    ----------
+    func : callable returning ndarray
+
+    Notes
+    -----
+    In general, graphicle tries to avoid memory copying. Instead, views
+    on the underlying data are used where possible. However, when
+    interfacing with a vendor which provides numpy arrays which are
+    column-contiguous, such as pandas, the assumed row-layout views will
+    fail. In these cases, the data will be copied into row-contiguous
+    arrays before returning. If already row-contiguous, the original
+    array is returned with no copying.
+    """
+
+    @fn.wraps(func)
+    def inner(*args, **kwargs):
+        return np.ascontiguousarray(func(*args, **kwargs))
+
+    return inner
+
+
+def _array_field(dtype: npt.DTypeLike, num_cols: int = 1):
+    """Abstracts out the dataclass field constructor for wrapped arrays,
+    providing standardised data cleaning, conversion, and comparison.
+
+    Parameters
+    ----------
+    dtype : dtype-like
+        Data type the underlying array should have. This should not be
+        a structured type, as these should be exposed to the user via
+        views.
+    num_cols : int
+        The number of columns the wrapped array will have.
+
+    Returns
+    -------
+    field : dataclass field
+    """
     dtype = np.dtype(dtype)
 
+    @_row_contiguous
     def converter(values: npt.ArrayLike) -> base.AnyVector:
         if isinstance(values, np.ndarray) and values.dtype.type == np.void:
             names = values.dtype.names
@@ -289,9 +371,10 @@ def _array_field(dtype: npt.DTypeLike, num_cols: int = 1, repr: bool = True):
             values = rfn.structured_to_unstructured(values)
         array = np.asarray(values, dtype=dtype)
         shape = array.shape
-        is_flat = len(shape) == 1
+        len_shape = len(shape)
+        is_flat = (len_shape == 1) or (len_shape == 0)
         if is_flat and (num_cols == 1):
-            return array
+            return array.reshape(-1)
         cols = shape[0] if is_flat else shape[1]
         not_empty = array.shape[0] != 0
         if not_empty and (cols != num_cols):
@@ -305,11 +388,12 @@ def _array_field(dtype: npt.DTypeLike, num_cols: int = 1, repr: bool = True):
         converter=converter,
         eq=cmp_using(np.array_equal),
         on_setattr=setters.convert,
-        repr=repr,
+        repr=False,
     )
 
 
 def _truthy(data: ty.Union[base.ArrayBase, base.AdjacencyBase]) -> bool:
+    """Defines the truthy value of the graphicle data structures."""
     if len(data) == 0:
         return False
     return True
@@ -318,12 +402,6 @@ def _truthy(data: ty.Union[base.ArrayBase, base.AdjacencyBase]) -> bool:
 ##################################
 # COMPOSITE MASK DATA STRUCTURES #
 ##################################
-class MaskAggOp(Enum):
-    AND = "and"
-    OR = "or"
-    NONE = None
-
-
 @define
 class MaskArray(base.MaskBase, base.ArrayBase):
     """Boolean mask over Graphicle data structures.
@@ -363,16 +441,18 @@ class MaskArray(base.MaskBase, base.ArrayBase):
 
     _data: base.BoolVector = _array_field("<?")
     dtype: np.dtype = field(init=False, repr=False)
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     _HANDLED_TYPES: ty.Tuple[ty.Type, ...] = field(init=False, repr=False)
 
     def __attrs_post_init__(self):
         self.dtype = self._data.dtype
-        self.__array_interface__ = self._data.__array_interface__
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     def __array__(self) -> base.BoolVector:
-        return self.data
+        return self._data
 
     @classmethod
     def __array_wrap__(cls, array: base.BoolVector) -> "MaskArray":
@@ -385,7 +465,7 @@ class MaskArray(base.MaskBase, base.ArrayBase):
         yield from map(bool, self.data)
 
     def copy(self) -> "MaskArray":
-        return deepcopy(self)
+        return self.__class__(self._data.copy())
 
     def __repr__(self) -> str:
         return _array_repr(self)
@@ -393,7 +473,7 @@ class MaskArray(base.MaskBase, base.ArrayBase):
     def __getitem__(self, key) -> "MaskArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     def __setitem__(self, key, val) -> None:
         if isinstance(key, base.MaskBase):
@@ -518,6 +598,15 @@ def _mask_dict_convert(masks: _IN_MASK_DICT) -> _MASK_DICT:
     return out_masks
 
 
+class MaskAggOp(Enum):
+    AND = "and"
+    OR = "or"
+    NONE = "none"
+
+
+AggStringLiterals = ty.Literal["or", "and", "none"]
+
+
 @define
 class MaskGroup(base.MaskBase, cla.MutableMapping[str, base.MaskBase]):
     """Data structure to compose groups of masks over particle arrays.
@@ -531,18 +620,17 @@ class MaskGroup(base.MaskBase, cla.MutableMapping[str, base.MaskBase]):
         Dictionary of MaskArray objects to be composed.
     agg_op : str or MaskAggOp
         Defines the aggregation operation when accessing the `data`
-        attribute. Options include "and" and "or".
-        Default is "and".
+        attribute. Options are "and", "or", "none". Default is "and".
 
     Attributes
     ----------
-    data : np.ndarray
+    data : ndarray[bool_]
         Combination of all masks in group via bitwise AND reduction.
     names : list[str]
         Provides the string values of the keys to the top-level nested
         ``MaskBase`` objects as a list. Will be deprecated in future.
         ``MaskGroup.keys()`` is preferred.
-    bitwise_or : np.ndarray[bool_]
+    bitwise_or : ndarray[bool_]
         Bitwise ``OR`` reduction over the nested masks.
     bitwise_or : np.ndarray[bool_]
         Bitwise ``AND`` reduction over the nested masks.
@@ -562,10 +650,21 @@ class MaskGroup(base.MaskBase, cla.MutableMapping[str, base.MaskBase]):
     _mask_arrays: _MASK_DICT = field(
         repr=False, factory=dict, converter=_mask_dict_convert
     )
-    agg_op: MaskAggOp = field(converter=MaskAggOp, default=MaskAggOp.AND)
+    _agg_op: AggStringLiterals = field(
+        default="and",
+        converter=lambda x: x.value if isinstance(x, MaskAggOp) else x,
+    )
+
+    @property
+    def agg_op(self) -> MaskAggOp:
+        return MaskAggOp(self._agg_op.lower())
+
+    @agg_op.setter
+    def agg_op(self, agg_op: AggStringLiterals) -> None:
+        self._agg_op = agg_op
 
     @classmethod
-    def from_numpy_structured(cls, arr: np.ndarray) -> "MaskGroup":
+    def from_numpy_structured(cls, arr: base.VoidVector) -> "MaskGroup":
         return cls(
             dict(  # type: ignore
                 map(lambda name: (name, arr[name]), arr.dtype.names)
@@ -614,7 +713,7 @@ class MaskGroup(base.MaskBase, cla.MutableMapping[str, base.MaskBase]):
             the ``MaskGroup`` whose components each individually have
             the passed slice applied.
         """
-        agg = self.agg_op
+        agg = self._agg_op
         if isinstance(key, list):
             return self.__class__(
                 cl.OrderedDict({k: self._mask_arrays[k] for k in key}), agg
@@ -789,22 +888,20 @@ class PdgArray(base.ArrayBase):
         Charge parity for each particle.
     """
 
-    from mcpid.lookup import PdgRecords as __PdgRecords
-
     _data: base.IntVector = _array_field("<i4")
     dtype: np.dtype = field(init=False, repr=False)
-    __lookup_table: __PdgRecords = field(init=False, repr=False)
     __mega_to_giga: float = field(init=False, repr=False)
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     _HANDLED_TYPES = field(init=False, repr=False)
 
     def __attrs_post_init__(self) -> None:
-        self.__lookup_table = self.__PdgRecords()
         self.__mega_to_giga: float = 1.0e-3
-        self.__array_interface__ = self._data.__array_interface__
         self.dtype = self._data.dtype
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     @classmethod
     def __array_wrap__(cls, array: base.IntVector) -> "PdgArray":
         return cls(array)
@@ -840,7 +937,7 @@ class PdgArray(base.ArrayBase):
     def __getitem__(self, key) -> "PdgArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     @property
     def data(self) -> base.IntVector:
@@ -851,7 +948,7 @@ class PdgArray(base.ArrayBase):
         self._data = values  # type: ignore
 
     def copy(self) -> "PdgArray":
-        return deepcopy(self)
+        return self.__class__(self._data)
 
     def mask(
         self,
@@ -891,12 +988,12 @@ class PdgArray(base.ArrayBase):
         )
 
     def __get_prop(self, field: str) -> base.VoidVector:
-        props = self.__lookup_table.properties(self.data, [field])[field]
+        props = _LOOKUP_TABLE.properties(self.data, [field])[field]
         return props  # type: ignore
 
     def __get_prop_range(self, field: str) -> base.VoidVector:
         field_range = [field + "lower", field + "upper"]
-        props = self.__lookup_table.properties(self.data, field_range)
+        props = _LOOKUP_TABLE.properties(self.data, field_range)
         return props  # type: ignore
 
     @property
@@ -972,6 +1069,14 @@ class MomentumArray(base.ArrayBase):
     ----------
     data : ndarray[float64]
         Structured array containing four momenta.
+    x : ndarray[float64]
+        x component of momentum.
+    y : ndarray[float64]
+        y component of momentum.
+    z : ndarray[float64]
+        z component of momentum.
+    energy : ndarray[float64]
+        Energy component of momentum.
     pt : ndarray[float64]
         Transverse component of particle momenta.
     rapidity : ndarray[float64]
@@ -992,16 +1097,18 @@ class MomentumArray(base.ArrayBase):
     """
 
     # data: base.AnyVector = array_field("pmu")
-    _data: base.DoubleVector = _array_field("<f8", num_cols=4, repr=False)
+    _data: base.DoubleVector = _array_field("<f8", num_cols=4)
     dtype: np.dtype = field(init=False, repr=False)
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     _HANDLED_TYPES: ty.Tuple[ty.Type, ...] = field(init=False, repr=False)
 
     def __attrs_post_init__(self):
         self.dtype = np.dtype(list(zip("xyze", it.repeat("<f8"))))
-        self.__array_interface__ = self._data.__array_interface__
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         return _array_ufunc(self, ufunc, method, *inputs, **kwargs)
 
@@ -1010,7 +1117,7 @@ class MomentumArray(base.ArrayBase):
 
     @property
     def data(self) -> base.VoidVector:
-        return self._data.view(self.dtype).squeeze()
+        return self._data.view(self.dtype).reshape(-1)
 
     @data.setter
     def data(self, values: npt.ArrayLike) -> None:
@@ -1026,7 +1133,7 @@ class MomentumArray(base.ArrayBase):
         yield from it.starmap(MomentumElement, elems)
 
     def __len__(self) -> int:
-        return len(self.data)
+        return self._data.shape[0]
 
     def __array__(self) -> base.VoidVector:
         return self.data
@@ -1034,7 +1141,7 @@ class MomentumArray(base.ArrayBase):
     def __getitem__(self, key) -> "MomentumArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     def __bool__(self) -> bool:
         return _truthy(self)
@@ -1050,28 +1157,52 @@ class MomentumArray(base.ArrayBase):
         return _array_ne(self, other)
 
     def copy(self) -> "MomentumArray":
-        return deepcopy(self)
+        return self.__class__(self._data)
 
     @property
     def _xy_pol(self) -> base.ComplexVector:
-        return self._data[:, :2].view(dtype="<c16").reshape(-1)
+        """Complex polar vector of momentum components in the x-y plane."""
+        return self._data[..., :2].view(dtype="<c16").reshape(-1)
 
     @fn.cached_property
     def _zt_pol(self) -> base.ComplexVector:
+        """Complex polar vector of momentum components in the
+        longitudinal-transverse plane.
+        """
         return (
-            np.stack((self.data["z"], np.abs(self._xy_pol)), axis=-1)
+            np.stack((self.z, np.abs(self._xy_pol)), axis=-1)
             .view(dtype="<c16")
             .reshape(-1)
         )
 
     @fn.cached_property
     def _spatial_mag(self) -> base.DoubleVector:
-        return np.abs(self._zt_pol)
+        return np.abs(self._zt_pol).reshape(-1)
+
+    @property
+    def x(self) -> base.DoubleVector:
+        """Momentum component along x-axis."""
+        return self.data["x"].reshape(-1)
+
+    @property
+    def y(self) -> base.DoubleVector:
+        """Momentum component along y-axis."""
+        return self.data["y"].reshape(-1)
+
+    @property
+    def z(self) -> base.DoubleVector:
+        """Momentum component along longitudinal / z-axis."""
+        return self.data["z"].reshape(-1)
+
+    @property
+    def energy(self) -> base.DoubleVector:
+        """Energy component of momentum."""
+        return self.data["e"].reshape(-1)
 
     @property
     def pt(self) -> base.DoubleVector:
         """Momentum component transverse to the beam-axis."""
-        return self._zt_pol.imag
+        return self._zt_pol.imag.reshape(-1)
 
     @fn.cached_property
     def eta(self) -> base.DoubleVector:
@@ -1083,34 +1214,33 @@ class MomentumArray(base.ArrayBase):
         """
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
-            arr = np.arctanh(self.data["z"] / self._spatial_mag)
-        return arr  # type: ignore
+            arr = np.arctanh(self.z / self._spatial_mag)
+        return arr.reshape(-1)
 
     @fn.cached_property
     def rapidity(self) -> base.DoubleVector:
         """Rapidity of particles."""
-        e, z = self.data["e"], self.data["z"]
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
-            rap = 0.5 * np.log((e + z) / (e - z))  # type: ignore
-        return rap  # type: ignore
+            rap = 0.5 * np.log((self.energy + self.z) / (self.energy - self.z))
+        return rap.reshape(-1)
 
     @fn.cached_property
     def phi(self) -> base.DoubleVector:
         """Azimuthal angular displacement of particles about beam-axis."""
-        return np.angle(self._xy_pol)
+        return np.angle(self._xy_pol).reshape(-1)
 
     @fn.cached_property
     def theta(self) -> base.DoubleVector:
         """Angular displacement of particles from positive beam-axis."""
-        return np.angle(self._zt_pol)
+        return np.angle(self._zt_pol).reshape(-1)
 
     @fn.cached_property
     def mass(self) -> base.DoubleVector:
         """Mass of particles."""
         return calculate._root_diff_two_squares(
-            self.data["e"], self._spatial_mag
-        )
+            self.energy, self._spatial_mag
+        ).reshape(-1)
 
     def delta_R(self, other: "MomentumArray") -> base.DoubleVector:
         """Calculates the Euclidean inter-particle distances in the
@@ -1170,17 +1300,19 @@ class ColorArray(base.ArrayBase):
     """
 
     _data: base.VoidVector = _array_field("<i4", 2)
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     dtype: np.dtype = field(init=False, repr=False)
     _HANDLED_TYPES: ty.Tuple[ty.Type, ...] = field(init=False, repr=False)
 
     def __attrs_post_init__(self):
-        self.__array_interface__ = self._data.__array_interface__
         self.dtype = np.dtype(
             list(zip(("color", "anticolor"), it.repeat("<i4")))
         )
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         return _array_ufunc(self, ufunc, method, *inputs, **kwargs)
 
@@ -1201,22 +1333,22 @@ class ColorArray(base.ArrayBase):
 
     @property
     def data(self) -> base.VoidVector:
-        return self._data.view(self.dtype).squeeze()
+        return self._data.view(self.dtype).reshape(-1)
 
     @data.setter
     def data(self, values: npt.ArrayLike) -> None:
         self._data = values  # type: ignore
 
-    def copy(self):
-        return deepcopy(self)
+    def copy(self) -> "ColorArray":
+        return self.__class__(self._data)
 
     def __getitem__(self, key) -> "ColorArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     def __len__(self) -> int:
-        return len(self.data)
+        return self._data.shape[0]
 
     def __bool__(self) -> bool:
         return _truthy(self)
@@ -1256,14 +1388,16 @@ class HelicityArray(base.ArrayBase):
 
     _data: base.HalfIntVector = _array_field("<i2")
     dtype: np.dtype = field(init=False, repr=False)
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     _HANDLED_TYPES: ty.Tuple[ty.Type, ...] = field(init=False, repr=False)
 
     def __attrs_post_init__(self):
-        self.__array_interface__ = self._data.__array_interface__
         self.dtype = self._data.dtype
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         return _array_ufunc(self, ufunc, method, *inputs, **kwargs)
 
@@ -1290,12 +1424,12 @@ class HelicityArray(base.ArrayBase):
 
     def copy(self) -> "HelicityArray":
         """Returns a new HelicityArray instance with same data."""
-        return deepcopy(self)
+        return self.__class__(self._data)
 
     def __getitem__(self, key) -> "HelicityArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     def __len__(self) -> int:
         return len(self.data)
@@ -1341,15 +1475,17 @@ class StatusArray(base.ArrayBase):
     """
 
     _data: base.HalfIntVector = _array_field("<i2")
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
     dtype: np.dtype = field(init=False, repr=False)
     _HANDLED_TYPES: ty.Tuple[ty.Type, ...] = field(init=False, repr=False)
 
     def __attrs_post_init__(self):
-        self.__array_interface__ = self._data.__array_interface__
         self.dtype = self._data.dtype
         self._HANDLED_TYPES = (np.ndarray, nm.Number, cla.Sequence)
 
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
+
     def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
         return _array_ufunc(self, ufunc, method, *inputs, **kwargs)
 
@@ -1369,7 +1505,7 @@ class StatusArray(base.ArrayBase):
     def __getitem__(self, key) -> "StatusArray":
         if isinstance(key, base.MaskBase):
             key = key.data
-        return self.__class__(np.array(self.data[key]))
+        return self.__class__(self._data[key])
 
     def __len__(self) -> int:
         return len(self.data)
@@ -1397,7 +1533,7 @@ class StatusArray(base.ArrayBase):
 
     def copy(self) -> "StatusArray":
         """Returns a new StatusArray instance with same data."""
-        return deepcopy(self)
+        return self.__class__(self._data)
 
     def in_range(
         self,
@@ -1444,7 +1580,7 @@ class StatusArray(base.ArrayBase):
                 "outgoing": data == 23,
                 "outgoing_nonperturbative_diffraction": data == 24,
             },
-            agg_op=MaskAggOp.OR,
+            agg_op="or",
         )
         return masks
 
@@ -1452,6 +1588,50 @@ class StatusArray(base.ArrayBase):
 #########################################
 # COMPOSITE OF PARTICLE DATA STRUCTURES #
 #########################################
+DsetPair = ty.Tuple[ty.Iterator[str], ty.Iterator[base.ArrayBase]]
+CompositeType = ty.Union["ParticleSet", "Graphicle"]
+CompositeGeneric = ty.TypeVar("CompositeGeneric", "ParticleSet", "Graphicle")
+
+
+def _dsets(instance: CompositeType) -> DsetPair:
+    names = instance.__annotations__.keys()
+    props = map(getattr, it.repeat(instance), names)
+    return iter(names), props
+
+
+def _nonempty_dsets(instance: CompositeType) -> DsetPair:
+    """Returns non-empty name and array iterators. Uses
+    ``itertools.tee()`` under the hood, so don't use if you intend
+    to iterate over them separately.
+    """
+    names, data = _dsets(instance)
+    data, data_ = it.tee(data)
+    pairs, pairs_ = it.tee(it.compress(zip(names, data), map(bool, data_)))
+    nonempty_names = map(op.itemgetter(0), pairs)
+    nonempty_data = map(op.itemgetter(1), pairs_)
+    return nonempty_names, nonempty_data
+
+
+def _composite_getitem(instance: CompositeGeneric, key) -> CompositeGeneric:
+    names, data = _nonempty_dsets(instance)
+    data_sliced = map(op.getitem, data, it.repeat(key))
+    return instance.__class__(**dict(zip(names, data_sliced)))
+
+
+def _composite_bool(instance: CompositeType) -> bool:
+    return any(map(bool, _dsets(instance)[1]))
+
+
+def _composite_len(instance: CompositeType) -> int:
+    return next(filter(fn.partial(op.lt, 0), map(len, _dsets(instance)[1])), 0)
+
+
+def _composite_copy(instance: CompositeGeneric) -> CompositeGeneric:
+    names, data = _nonempty_dsets(instance)
+    copies = map(op.methodcaller("copy"), data)
+    return instance.__class__(**dict(zip(names, copies)))
+
+
 @define
 class ParticleSet(base.ParticleBase):
     """Composite of data structures containing particle set description.
@@ -1489,45 +1669,29 @@ class ParticleSet(base.ParticleBase):
         Boolean array indicating final state in particle set.
     """
 
-    pdg: PdgArray = field(default=Factory(fn.partial(PdgArray, tuple())))
-    pmu: MomentumArray = field(default=Factory(fn.partial(PdgArray, tuple())))
-    color: ColorArray = ColorArray()
-    helicity: HelicityArray = HelicityArray()
-    status: StatusArray = StatusArray()
-    final: MaskArray = MaskArray()
-
-    @property
-    def __dsets(self):
-        for dset_name in tuple(self.__annotations__.keys()):
-            yield {"name": dset_name, "data": getattr(self, dset_name)}
+    pdg: PdgArray = field(default=Factory(PdgArray))
+    pmu: MomentumArray = field(default=Factory(MomentumArray))
+    color: ColorArray = field(default=Factory(ColorArray))
+    helicity: HelicityArray = field(default=Factory(HelicityArray))
+    status: StatusArray = field(default=Factory(StatusArray))
+    final: MaskArray = field(default=Factory(MaskArray))
 
     def __getitem__(self, key) -> "ParticleSet":
-        kwargs = dict()
-        for dset in self.__dsets:
-            name = dset["name"]
-            data = dset["data"]
-            if len(data) > 0:
-                kwargs.update({name: data[key]})
-        return self.__class__(**kwargs)
+        return _composite_getitem(self, key)
 
     def __bool__(self) -> bool:
-        for dset in self.__dsets:
-            if dset["data"]:
-                return True
-        return False
+        return _composite_bool(self)
 
     def __repr__(self) -> str:
-        dset_repr = (repr(dset["data"]) for dset in self.__dsets)
-        dset_str = ",\n".join(dset_repr)
+        data = _dsets(self)[1]
+        dset_str = ",\n".join(map(repr, data))
         return f"ParticleSet(\n{dset_str}\n)"
 
     def __len__(self) -> int:
-        filled_dsets = filter(lambda dset: len(dset["data"]) > 0, self.__dsets)
-        dset = next(filled_dsets)
-        return len(dset)
+        return _composite_len(self)
 
     def copy(self) -> "ParticleSet":
-        return deepcopy(self)
+        return _composite_copy(self)
 
     @classmethod
     def from_numpy(
@@ -1617,10 +1781,10 @@ class AdjacencyList(base.AdjacencyBase):
 
     _data: base.AnyVector = array_field("edge")
     weights: base.DoubleVector = array_field("double")
-    __array_interface__: ty.Dict[str, ty.Any] = field(init=False, repr=False)
 
-    def __attrs_post_init__(self):
-        self.__array_interface__ = self._data.__array_interface__
+    @property
+    def __array_interface__(self) -> ty.Dict[str, ty.Any]:
+        return self._data.__array_interface__
 
     @classmethod
     def __array_wrap__(cls, array: base.AnyVector) -> "AdjacencyList":
@@ -1838,29 +2002,20 @@ class Graphicle:
         Vertex at which the hard process is initiated.
     """
 
-    particles: ParticleSet = ParticleSet()
-    adj: AdjacencyList = AdjacencyList()
-
-    @property
-    def __attr_names(self) -> ty.Tuple[str, ...]:
-        return tuple(self.__annotations__.keys())
+    particles: ParticleSet = field(default=Factory(ParticleSet))
+    adj: AdjacencyList = field(default=Factory(AdjacencyList))
 
     def __getitem__(self, key) -> "Graphicle":
-        kwargs = dict()
-        for name in self.__attr_names:
-            data = getattr(self, name)
-            if len(data) != 0:
-                kwargs.update({name: data[key]})
-        return self.__class__(**kwargs)
+        return _composite_getitem(self, key)
 
     def __bool__(self) -> bool:
-        for name in self.__attr_names:
-            if getattr(self, name):
-                return True
-        return False
+        return _composite_bool(self)
+
+    def __len__(self) -> int:
+        return _composite_len(self)
 
     def copy(self) -> "Graphicle":
-        return deepcopy(self)
+        return _composite_copy(self)
 
     @classmethod
     def from_event(cls, event: base.EventInterface) -> "Graphicle":
